import json
from typing import TYPE_CHECKING

from django import forms
from django.utils.functional import cached_property, classproperty  # type: ignore
from wagtail.admin.panels import (
    FieldPanel,
    MultipleChooserPanel,
    Panel,
    TitleFieldPanel,
)
from wagtail.admin.staticfiles import versioned_static

from wagtail_ai.agents.basic_prompt import PageDescriptionPrompt, PageTitlePrompt

if TYPE_CHECKING:
    from django_ai_core.contrib.index import VectorIndex


class AIPanelMixin(Panel):
    def __init__(self, *args, prompts=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.prompts = prompts or []
        self.attrs["data-wai-field-panel-prompts-value"] = json.dumps(self.prompts)

    @classproperty
    def BASE_ATTRS(cls):
        base = super().BASE_ATTRS
        controllers = [base.get("data-controller", ""), "wai-field-panel"]
        return {
            **base,
            "data-controller": " ".join(controllers).strip(),
        }

    def clone_kwargs(self):
        kwargs = super().clone_kwargs()
        kwargs["prompts"] = self.prompts
        return kwargs

    class BoundPanel(Panel.BoundPanel):
        @cached_property
        def media(self):  # type: ignore
            return super().media + forms.Media(
                js=[versioned_static("wagtail_ai/field_panel.js")],
                css={"all": [versioned_static("wagtail_ai/field_panel.css")]},
            )


class AIFieldPanel(AIPanelMixin, FieldPanel):
    class BoundPanel(AIPanelMixin.BoundPanel, FieldPanel.BoundPanel):  # type: ignore
        pass


class AITitleFieldPanel(AIPanelMixin, TitleFieldPanel):
    def __init__(self, *args, prompts=None, **kwargs):
        if prompts is None:
            prompts = [PageTitlePrompt.name]
        super().__init__(*args, prompts=prompts, **kwargs)

    class BoundPanel(AIPanelMixin.BoundPanel, TitleFieldPanel.BoundPanel):  # type: ignore
        pass


class AIDescriptionFieldPanel(AIFieldPanel):
    def __init__(self, *args, prompts=None, **kwargs):
        if prompts is None:
            prompts = [PageDescriptionPrompt.name]
        super().__init__(*args, prompts=prompts, **kwargs)


class AIChooserPanelMixin(Panel):
    def __init__(
        self,
        *args,
        suggest_limit: int = 3,
        chunk_size: int | None = None,
        vector_index: "str | VectorIndex",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.suggest_limit = suggest_limit
        self.chunk_size = chunk_size
        if isinstance(vector_index, str):
            self.vector_index = vector_index
        else:
            self.vector_index = vector_index.index_id

    def clone_kwargs(self):
        return {
            **super().clone_kwargs(),
            "vector_index": self.vector_index,
            "suggest_limit": self.suggest_limit,
            "chunk_size": self.chunk_size,
        }

    @classproperty
    def BASE_ATTRS(cls):
        base = super().BASE_ATTRS
        controllers = [base.get("data-controller", ""), "wai-chooser-panel"]
        actions = [
            base.get("data-action", ""),
            "w-formset:removed->wai-chooser-panel#updateControlStates",
            "w-formset:added->wai-chooser-panel#updateControlStates",
            "w-formset:ready->wai-chooser-panel#updateControlStates",
        ]
        return {
            **base,
            "data-controller": " ".join(controllers).strip(),
            "data-action": " ".join(actions).strip(),
        }

    class BoundPanel(Panel.BoundPanel):
        template_name = "wagtail_ai/panels/chooser_panel.html"

        @classproperty
        def original_template_name(cls):
            return super().template_name

        @property
        def attrs(self):
            attrs = super().attrs
            attrs["data-wai-chooser-panel-relation-name-value"] = (
                self.panel.relation_name
            )
            attrs["data-wai-chooser-panel-instance-pk-value"] = (
                self.instance.pk if self.instance else None
            )
            attrs["data-wai-chooser-panel-limit-value"] = self.panel.suggest_limit
            attrs["data-wai-chooser-panel-vector-index-value"] = self.panel.vector_index
            if self.panel.chunk_size:
                attrs["data-wai-chooser-panel-chunk-size-value"] = self.panel.chunk_size
            return attrs

        def get_context_data(self, parent_context=None):
            context = super().get_context_data(parent_context=parent_context) or {}
            context["original_template_name"] = self.original_template_name
            return context

        @cached_property
        def media(self):  # type: ignore
            return super().media + forms.Media(
                js=[versioned_static("wagtail_ai/chooser_panel.js")],
                css={"all": [versioned_static("wagtail_ai/chooser_panel.css")]},
            )


class AIMultipleChooserPanel(AIChooserPanelMixin, MultipleChooserPanel):
    class BoundPanel(AIChooserPanelMixin.BoundPanel, MultipleChooserPanel.BoundPanel):  # type: ignore
        pass



import json
from functools import cached_property

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from wagtail.admin.staticfiles import versioned_static
from wagtail.images.forms import BaseImageForm

from wagtail_ai.agents.basic_prompt import ImageDescriptionPrompt, ImageTitlePrompt
from wagtail_ai.models import Prompt


class PromptTextField(forms.CharField):
    default_error_messages = {
        "required": _(
            "No text provided - please enter some text before using AI features."
        ),
    }


class PromptUUIDField(forms.UUIDField):
    default_error_messages = {
        "required": _("Invalid prompt provided."),
        "invalid": _("Invalid prompt provided."),
    }


class ApiForm(forms.Form):
    def errors_for_json_response(self) -> str:
        errors_for_response = []
        for _field, errors in self.errors.get_json_data().items():
            for error in errors:
                errors_for_response.append(error["message"])

        return " \n".join(errors_for_response)


class PromptForm(ApiForm):
    text = PromptTextField()
    prompt = PromptUUIDField()

    def clean_prompt(self):
        prompt_uuid = self.cleaned_data["prompt"]
        prompt = Prompt.objects.filter(uuid=prompt_uuid).first()
        if not prompt:
            raise ValidationError(
                self.fields["prompt"].error_messages["invalid"], code="invalid"
            )

        return prompt


class DescribeImageApiForm(ApiForm):
    image_id = forms.CharField()
    maxlength = forms.IntegerField(required=False, min_value=0, max_value=4096)


class ImageDescriptionWidgetMixin(forms.Widget):
    """
    A widget mixin that wraps the widget in a controller for describing images.
    Can be used with TextInput or Textarea.
    """

    def __init__(
        self,
        *args,
        image_id=None,
        file_selector=None,
        prompts=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.attrs["data-controller"] = " ".join(
            {
                *self.attrs.get("data-controller", "").split(),
                "wai-field-panel",
            }
        )
        if image_id:
            self.attrs["data-wai-field-panel-image-id"] = image_id
        self.attrs["data-wai-field-panel-image-input-value"] = file_selector
        self.attrs["data-wai-field-panel-prompts-value"] = json.dumps(
            prompts or [ImageDescriptionPrompt.name]
        )

    @cached_property
    def media(self):  # type: ignore
        return forms.Media(
            js=[
                versioned_static("wagtail_ai/field_panel.js"),
            ],
            css={
                "all": [versioned_static("wagtail_ai/field_panel.css")],
            },
        )


class ImageDescriptionTextInput(ImageDescriptionWidgetMixin, forms.TextInput):
    pass


class DescribeImageForm(BaseImageForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        widget_kwargs = {}
        if "file" in self.fields:
            widget_kwargs["file_selector"] = f"#{self['file'].id_for_label}"
        if self.instance and self.instance.pk:
            widget_kwargs["image_id"] = self.instance.pk

        self.fields["title"].widget = ImageDescriptionTextInput(
            **widget_kwargs,
            prompts=[ImageTitlePrompt.name],
            attrs=self.fields["title"].widget.attrs,
        )
        self.fields["description"].widget = ImageDescriptionTextInput(
            **widget_kwargs,
            prompts=[ImageDescriptionPrompt.name],
            attrs=self.fields["description"].widget.attrs,
        )



from django_ai_core.contrib.agents import (
    Agent,
    AgentParameter,
)
from django_ai_core.contrib.agents import (
    registry as agent_registry,
)
from django_ai_core.contrib.index import registry as index_registry
from django_ai_core.contrib.index.chunking import SimpleChunkTransformer
from wagtail.admin.admin_url_finder import AdminURLFinder

MAX_LIMIT = 100


@agent_registry.register()
class SuggestedContentAgent(Agent):
    slug = "wai_suggested_content"
    description = "Suggests similar content to the provided context using the provided VectorIndex"
    parameters = [
        AgentParameter(
            name="vector_index",
            type=str,
            description="ID of the VectorIndex to query",
        ),
        AgentParameter(
            name="exclude_pks",
            type=list[str],
            description="PKs to exclude from results",
        ),
        AgentParameter(
            name="content",
            type=str,
            description="Content to use in similarity search",
        ),
        AgentParameter(
            name="limit",
            type=int,
            description="Number of responses to return",
        ),
        AgentParameter(
            name="chunk_size",
            type=int,
            description="Number of characters to embed from content",
        ),
    ]

    def execute(
        self,
        vector_index: str,
        exclude_pks: list[str],
        content: str,
        limit: int = 3,
        chunk_size: int = 1000,
    ) -> list:
        index_cls = index_registry.get(vector_index)
        index = index_cls()
        finder = AdminURLFinder()
        # Extend limit by the number of excluded PKs so we get enough responses
        extended_limit = limit + len(exclude_pks)

        if extended_limit > MAX_LIMIT:
            return []

        if not chunk_size:
            chunk_size = 1000

        chunk_transformer = SimpleChunkTransformer(chunk_size=chunk_size)
        chunks = chunk_transformer.transform(content)

        if not chunks:
            return []

        return [
            {
                "id": str(page.pk),
                "title": page.title,
                "editUrl": finder.get_edit_url(page),
            }
            for page in index.search_sources(chunks[0])[:extended_limit]
            if str(page.pk) not in exclude_pks
        ][:limit]



import warnings
from functools import cache
from typing import TYPE_CHECKING, cast

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.dispatch import receiver
from django.test.signals import setting_changed
from django.utils.module_loading import import_string
from django_ai_core.llm import LLMService
from wagtail.contrib.settings.models import BaseGenericSetting, BaseSiteSetting
from wagtail.models import Site

from wagtail_ai import ai
from wagtail_ai.utils.deprecation import WagtailAISettingsDeprecationWarning

if TYPE_CHECKING:
    from wagtail_ai.models import AgentSettingsMixin

_DEFAULT_MODEL_PROVIDER = "openai"
DEFAULT_PROVIDER_ALIAS = "default"


def get_providers() -> dict[str, dict[str, str]]:
    return getattr(settings, "WAGTAIL_AI", {}).get("PROVIDERS", {})


def get_provider(alias=DEFAULT_PROVIDER_ALIAS) -> dict[str, str]:
    """
    Get a provider configuration by alias.

    Args:
        alias: The provider alias to look up

    Returns:
        Provider configuration dict with ``"provider"`` and ``"model"`` keys,
        and any other keys to be passed as keyword arguments to the ``LLMService`` constructor.

    Raises:
        ``ImproperlyConfigured``: If provider not found
    """
    providers = get_providers()
    provider = providers.get(alias)

    # Legacy fallback for backwards compatibility (only for DEFAULT_PROVIDER_ALIAS)
    if not provider and alias == DEFAULT_PROVIDER_ALIAS:
        backend = ai.get_backend()
        model = backend.config.model_id
        warnings.warn(
            f"No '{DEFAULT_PROVIDER_ALIAS}' provider configured in "
            "WAGTAIL_AI['PROVIDERS'], falling back to "
            f"'{_DEFAULT_MODEL_PROVIDER}' provider with the '{model}' model from "
            "the default backend's 'MODEL_ID' setting. This fallback behavior is "
            "deprecated and will be removed in a future release. "
            "Please configure a default provider in WAGTAIL_AI['PROVIDERS'].",
            WagtailAISettingsDeprecationWarning,
            stacklevel=2,
        )
        provider = {"provider": _DEFAULT_MODEL_PROVIDER, "model": model}
        return provider

    if not provider:
        raise ImproperlyConfigured(
            f"Provider with alias '{alias}' not found in WAGTAIL_AI['PROVIDERS']."
        )

    return provider


@cache
def get_llm_service(alias=DEFAULT_PROVIDER_ALIAS) -> LLMService:
    """
    Get an LLM service instance for the given provider alias.

    Args:
        alias: The provider alias to look up

    Returns:
        LLMService instance configured with the specified provider
    """
    return LLMService.create(**get_provider(alias))


def get_agent_settings_model() -> "type[BaseGenericSetting | BaseSiteSetting]":
    wagtail_ai_settings = getattr(settings, "WAGTAIL_AI", {})
    model_path = wagtail_ai_settings.get(
        "AGENT_SETTINGS_MODEL",
        "wagtail_ai.models.AgentSettings",
    )
    return import_string(model_path)


def get_agent_settings() -> "AgentSettingsMixin":
    model = get_agent_settings_model()
    if issubclass(model, BaseSiteSetting):
        # TODO: support multiple sites
        settings = model.for_site(site=Site.objects.first())
    else:
        settings = model.load()

    return cast("AgentSettingsMixin", settings)


@receiver(setting_changed)
def clear_caches_on_setting_change(sender, setting, **kwargs):
    if setting == "WAGTAIL_AI":
        get_llm_service.cache_clear()



import json

from django import forms
from django.utils.functional import cached_property
from wagtail.admin.staticfiles import versioned_static
from wagtail.admin.telepath import register
from wagtail.blocks.struct_block import StructBlockAdapter
from wagtail.images.blocks import ImageBlock

from wagtail_ai.agents.basic_prompt import ContextualAltTextPrompt


def get_image_block_form_attrs(
    alt_text_field_name: str = "alt_text",
    image_field_name: str = "image",
    prompts=None,
):
    prompts = prompts or [ContextualAltTextPrompt.name]
    return {
        "data-controller": "wai-field-panel",
        "data-wai-field-panel-main-input-value": f'[data-contentpath="{alt_text_field_name}"] input',
        "data-wai-field-panel-image-input-value": f'[data-contentpath="{image_field_name}"] input',
        "data-wai-field-panel-prompts-value": json.dumps(prompts),
    }


class AIImageBlockAdapter(StructBlockAdapter):
    js_constructor = "wagtail_ai.blocks.AIImageBlock"

    @cached_property
    def media(self):
        return super().media + forms.Media(
            js=[
                versioned_static("wagtail_ai/field_panel.js"),
                versioned_static("wagtail_ai/image_block.js"),
            ],
            css={"all": [versioned_static("wagtail_ai/field_panel.css")]},
        )


def ai_image_block(
    alt_text_field_name: str = "alt_text",
    image_field_name: str = "image",
    prompts=None,
):
    """Decorator for a block class to set the appropriate form attributes for AI image features."""

    def decorator(block_class):
        block_class._meta_class.form_attrs = {
            **(block_class._meta_class.form_attrs or {}),
            **get_image_block_form_attrs(
                alt_text_field_name=alt_text_field_name,
                image_field_name=image_field_name,
                prompts=prompts,
            ),
        }
        register(AIImageBlockAdapter(), block_class)
        return block_class

    return decorator


@ai_image_block()
class AIImageBlock(ImageBlock):
    pass



import base64
from abc import ABC
from urllib.parse import SplitResult

from django.conf import settings
from django.core.files import File
from django.utils.translation import gettext_lazy as _
from django_ai_core.contrib.agents import Agent, AgentParameter, registry

from wagtail_ai.context import PromptContext

from .base import get_llm_service

# Temporary drop-in for the Prompt model using prompts from AgentSettings


class SettingPrompt(ABC):
    name: str
    label: str
    description: str


class PageTitlePrompt(SettingPrompt):
    name = "page_title_prompt"
    label = _("Generate title")
    description = _("Generate a title based on the page content")


class PageDescriptionPrompt(SettingPrompt):
    name = "page_description_prompt"
    label = _("Generate description")
    description = _("Generate a description by summarizing the page content")


class ImageTitlePrompt(SettingPrompt):
    name = "image_title_prompt"
    label = _("Generate title")
    description = _("Generate a title for the image")


class ImageDescriptionPrompt(SettingPrompt):
    name = "image_description_prompt"
    label = _("Generate description")
    description = _("Generate a description for the image")


class ContextualAltTextPrompt(SettingPrompt):
    name = "contextual_alt_text_prompt"
    label = _("Generate alt text")
    description = _("Generate contextual alt text for the image")


@registry.register()
class BasicPromptAgent(Agent):
    slug = "wai_basic_prompt"
    description = (
        "Provides suggestions to an input field based on a given prompt and context."
    )
    parameters = [
        AgentParameter(
            name="prompt",
            type=str,
            description="The prompt to guide the suggestion.",
        ),
        AgentParameter(
            name="context",
            type=dict[str, str],
            description="Additional context to replace placeholders in the prompt text.",
        ),
    ]
    settings_prompts = [
        PageTitlePrompt,
        PageDescriptionPrompt,
        ImageTitlePrompt,
        ImageDescriptionPrompt,
        ContextualAltTextPrompt,
    ]
    provider_alias = "default"

    def execute(self, prompt: str, context: dict[str, str]) -> str:
        self.prompt = prompt
        self.context = self.validate_context(prompt, context)
        if self.context.get("image"):
            ai_settings = getattr(settings, "WAGTAIL_AI", {})
            self.provider_alias = ai_settings.get(
                "IMAGE_DESCRIPTION_PROVIDER", "default"
            )

        messages = self._get_prompt_messages()
        result = self._get_result(messages)
        return result

    def _get_prompt_messages(self):
        files = self.split_context_files()
        messages = [
            {
                "role": "user",
                "content": [
                    {
                        "type": "text",
                        "text": self.get_formatted_prompt(),
                    },
                    *files,
                ],
            },
        ]
        return messages

    def _get_result(self, messages: list[dict]) -> str:
        llm_service = get_llm_service(alias=self.provider_alias)
        result = llm_service.completion(messages=messages)
        return result.choices[0].message.content  # type: ignore

    def validate_context(self, prompt: str, context: dict[str, str]) -> PromptContext:
        context = PromptContext(context)
        context.clean(prompt)
        return context

    def get_formatted_prompt(self) -> str:
        return self.prompt.format_map(self.context)

    def split_context_files(self) -> list[File]:
        files = []
        for key, value in self.context.items():
            if isinstance(value, (File, SplitResult)):
                if isinstance(value, File):
                    # TODO: check the file type and handle accordingly
                    with value.open() as f:
                        base64_image = base64.b64encode(f.read()).decode()
                    url = f"data:image/jpeg;base64,{base64_image}"
                else:
                    # Assume it's a data URL for an image
                    url = value.geturl()
                files.append(
                    {
                        "type": "image_url",
                        "image_url": {"url": url},
                    }
                )
                self.context[key] = f"[file {len(files)}]"
        return files



import json
from enum import IntEnum

from django_ai_core.contrib.agents import Agent, AgentParameter, registry
from pydantic import BaseModel, Field

from .base import get_agent_settings, get_llm_service


class QualityScore(IntEnum):
    NEEDS_MAJOR_IMPROVEMENT = 1
    ADEQUATE = 2
    EXCELLENT = 3


class SpecificImprovement(BaseModel):
    original_text: str = Field(description="The original text that needs improvement")
    suggested_text: str = Field(
        description=(
            "The suggested revised text in the content's language, in plain text"
        )
    )
    explanation: str = Field(
        description=(
            "A brief explanation in plain text in the editor's language for why "
            "this change would improve the content"
        )
    )


class ContentFeedbackSchema(BaseModel):
    quality_score: QualityScore = Field(
        description=(
            "Quality score between 1 and 3 "
            "(1=needs major improvement, 2=adequate, 3=excellent)"
        ),
    )
    qualitative_feedback: list[str] = Field(
        description=(
            "3-5 bullet points of qualitative feedback highlighting strengths "
            "and areas for improvement"
        ),
        min_length=3,
        max_length=5,
    )
    specific_improvements: list[SpecificImprovement] = Field(
        description=(
            "Specific text improvements with original text, suggested revised text, "
            "and a brief explanation for why each change would improve the content"
        ),
        min_length=1,
    )


@registry.register()
class ContentFeedbackAgent(Agent):
    slug = "wai_content_feedback"
    description = "An agent that provides feedback on content quality."
    parameters = [
        AgentParameter(
            name="content_text",
            type=str,
            description="The text content to review",
        ),
        AgentParameter(
            name="content_html",
            type=str,
            description="The HTML content to review",
        ),
        AgentParameter(
            name="content_language",
            type=str,
            description="The language of the content (e.g. 'American English')",
        ),
        AgentParameter(
            name="editor_language",
            type=str,
            description=(
                "The preferred language for the editor interface (e.g. 'French')"
            ),
        ),
    ]
    provider_alias = "default"
    _response_format = ContentFeedbackSchema

    def execute(
        self,
        content_text: str,
        content_html: str,
        content_language: str,
        editor_language: str,
    ) -> dict:
        settings = get_agent_settings()
        messages = [
            {
                "role": "system",
                "content": "You are a helpful assistant that responds with structured data according to the provided schema. The language rules specified are IMPORTANT. Always ensure the feedback and improvements are in the correct language.",
            },
            {
                "role": "system",
                "content": f"""Analyze the given content and provide:
1. A quality score between 1 and 3 (1=needs major improvement, 2=adequate, 3=excellent)
2. 3-5 bullet points of qualitative feedback highlighting strengths and areas for improvement in {editor_language}
3. Specific text improvements with original text, suggested revised text in {content_language}, and a brief explanation
   in {editor_language} for why each change would improve the content. Provide text
   improvements in plain text, not HTML and not markdown.

The language rules specified are IMPORTANT. Always ensure the feedback and improvements are in the correct language.

Return JSON with the provided structure WITHOUT the markdown code block. Start immediately with a {{ character and end with a }} character.""",
            },
        ]

        messages += self._get_prompt_messages(settings)

        match settings.content_feedback_content_type:
            case settings.ContentFeedbackContentType.TEXT:
                content = content_text
            case settings.ContentFeedbackContentType.HTML | _:
                content = content_html

        messages.append(
            {
                "role": "system",
                "content": f"Content to review:\n\n{content}",
            }
        )

        return self._get_result(messages)

    def _get_prompt_messages(self, settings) -> list[dict]:
        messages = []
        if settings.content_feedback_prompt:
            messages.append(
                {
                    "role": "user",
                    "content": settings.content_feedback_prompt,
                }
            )
        return messages

    def _get_result(self, messages: list[dict]) -> dict:
        client = get_llm_service(alias=self.provider_alias)
        result = client.completion(
            messages=messages,
            response_format=self._response_format,
        )
        return json.loads(result.choices[0].message.content)  # type: ignore



import uuid
from typing import NotRequired, Required, TypedDict, cast

from django.forms.utils import flatatt
from django.template.loader import render_to_string
from django.urls import include, path, reverse
from django.utils.html import format_html, json_script
from django.views.i18n import JavaScriptCatalog
from django_ai_core.contrib.agents import registry
from wagtail import hooks
from wagtail.admin.rich_text.editors.draftail.features import ControlFeature
from wagtail.admin.staticfiles import versioned_static
from wagtail.contrib.settings.models import register_setting

from wagtail_ai.agents.base import get_agent_settings, get_agent_settings_model
from wagtail_ai.agents.basic_prompt import BasicPromptAgent

from .models import Prompt
from .views import describe_image, prompt_viewset, text_completion


@hooks.register("register_admin_urls")  # type: ignore
def register_admin_urls():
    content_feedback_agent = registry.get("wai_content_feedback")
    basic_prompt_agent = registry.get("wai_basic_prompt")
    suggested_content_agent = registry.get("wai_suggested_content")

    urls = [
        path(
            "jsi18n/",
            JavaScriptCatalog.as_view(packages=["wagtail_ai"]),
            name="javascript_catalog",
        ),
        path(
            "text_completion/",
            text_completion,
            name="text_completion",
        ),
        path(
            "describe_image/",
            describe_image,
            name="describe_image",
        ),
        path(
            "content_feedback/",
            content_feedback_agent.as_view(),
            name="content_feedback",
        ),
        path(
            "basic_prompt/",
            basic_prompt_agent.as_view(),
            name="basic_prompt",
        ),
        path(
            "suggested_content/",
            suggested_content_agent.as_view(),
            name="suggested_content",
        ),
    ]

    return [
        path(
            "ai/",
            include(
                (urls, "wagtail_ai"),
                namespace="wagtail_ai",
            ),
        )
    ]


@hooks.register("register_rich_text_features")  # type: ignore
def register_ai_feature(features):
    feature_name = "ai"
    features.default_features.append(feature_name)

    features.register_editor_plugin(
        "draftail",
        feature_name,
        ControlFeature(
            {
                "type": feature_name,
            },
            js=["wagtail_ai/draftail.js"],
            css={"all": ["wagtail_ai/draftail.css"]},
        ),
    )


class PromptDict(TypedDict):
    # Fields should match the Prompt type defined in custom.d.ts
    # be careful not to expose any sensitive or exploitable data here.
    uuid: Required[uuid.UUID]
    default_prompt_id: NotRequired[int | None]
    label: Required[str]
    description: NotRequired[str]
    prompt: Required[str]
    method: Required[str]


def _serialize_prompt(prompt: Prompt) -> PromptDict:
    return {
        "uuid": prompt.uuid,
        "default_prompt_id": prompt.default_prompt_id,
        "label": prompt.label,
        "description": prompt.description,
        "prompt": prompt.prompt_value,
        "method": prompt.method,
    }


def get_prompts():
    return [_serialize_prompt(prompt) for prompt in Prompt.objects.all()]


def get_setting_prompts():
    settings = get_agent_settings()
    basic_prompt_agent = cast(
        type[BasicPromptAgent],
        registry.get("wai_basic_prompt"),
    )
    return [
        {
            "name": prompt.name,
            "label": prompt.label,
            "description": prompt.description,
            "prompt": getattr(settings, prompt.name),
        }
        for prompt in basic_prompt_agent.settings_prompts
    ]


@hooks.register("insert_global_admin_css")  # type: ignore
def ai_admin_css():
    return format_html(
        '<link rel="stylesheet" href="{}">', versioned_static("wagtail_ai/main.css")
    )


@hooks.register("insert_global_admin_js")  # type: ignore
def ai_admin_js():
    config = {
        "aiPrompts": get_prompts(),
        "settingPrompts": get_setting_prompts(),
        "urls": {
            "TEXT_COMPLETION": reverse("wagtail_ai:text_completion"),
            "DESCRIBE_IMAGE": reverse("wagtail_ai:describe_image"),
            "CONTENT_FEEDBACK": reverse("wagtail_ai:content_feedback"),
            "BASIC_PROMPT": reverse("wagtail_ai:basic_prompt"),
            "SUGGESTED_CONTENT": reverse("wagtail_ai:suggested_content"),
        },
    }

    return format_html(
        '{}<script src="{}"></script><script src="{}"></script>',
        json_script(config, "wagtail-ai-config"),
        versioned_static("wagtail_ai/main.js"),
        reverse("wagtail_ai:javascript_catalog"),
    )


@hooks.register("insert_global_admin_js")  # type: ignore
def ai_editor_js():
    dropdown_attrs = {
        "data-wai-field-panel-target": "dropdown",
    }
    context = {
        "dropdown_attrs": flatatt(dropdown_attrs),
    }
    return render_to_string("wagtail_ai/shared/field_panel_dropdown.html", context)


@hooks.register("register_admin_viewset")  # type: ignore
def register_viewset():
    return prompt_viewset


@hooks.register("register_icons")  # type: ignore
def register_icons(icons):
    return [
        *icons,
        "wagtail_ai/icons/wand.svg",
        "wagtail_ai/icons/wand-animated.svg",
    ]


register_setting(get_agent_settings_model(), order=prompt_viewset.menu_order - 1)



import uuid

from django.db import models
from django.utils.translation import gettext_lazy as _
from wagtail.admin.panels import FieldPanel, MultiFieldPanel
from wagtail.contrib.settings.models import BaseGenericSetting
from wagtail.search import index

from wagtail_ai.prompts import DEFAULT_PROMPTS


class Prompt(models.Model, index.Indexed):
    class Method(models.TextChoices):
        REPLACE = "replace", _("Replace content")
        APPEND = "append", _("Append after existing content")

    uuid = models.UUIDField(default=uuid.uuid4, unique=True, editable=False)
    default_prompt_id = models.SmallIntegerField(unique=True, editable=False, null=True)
    label = models.CharField(max_length=50)
    description = models.CharField(
        max_length=255,
        blank=True,
        help_text=_(
            "The prompt description, displayed alongside its label on the settings listing."
        ),
    )
    prompt = models.TextField(  # noqa: DJ001
        null=True,  # field is nullable to manage default prompts, see the prompt_value property for more info.
        blank=False,
        help_text=_(
            "The prompt text sent to the Large Language Model (e.g. ChatGPT) to generate content."
        ),
    )
    method = models.CharField(
        max_length=25,
        choices=Method.choices,
        help_text=_("The method used for processing the responses to the prompt."),
    )

    search_fields = [
        index.AutocompleteField("label"),
        index.SearchField("description"),
        index.SearchField("prompt"),
    ]

    def __str__(self):
        return self.label

    def get_default_prompt_value(self) -> str:
        """
        Return the default prompt value from DEFAULT_PROMPTS.
        """
        return next(
            (
                prompt["prompt"]
                for prompt in DEFAULT_PROMPTS
                if prompt["default_prompt_id"] == self.default_prompt_id
            ),
            "",
        )

    @property
    def is_default(self) -> bool:
        """
        Returns True if the prompt is one of the default prompts.
        """
        return self.default_prompt_id is not None

    @property
    def prompt_value(self) -> str:
        """
        Return the prompt value, otherwise if the prompt is None and belongs
        to the default prompts, map to the default prompt value.
        """
        if not self.prompt:
            if self.is_default:
                return self.get_default_prompt_value()
            else:
                raise ValueError("Prompt value is empty and not a default prompt.")
        return self.prompt


class AgentPromptDefaults:
    @classmethod
    def page_title_prompt(cls):
        return (
            "Create an SEO-friendly page title, and respond ONLY with the title "
            "in plain text (without quotes), for the following "
            "web page content:\n\n"
            "{content_html}"
        )

    @classmethod
    def page_description_prompt(cls):
        return (
            "Create an SEO-friendly meta description, and respond ONLY with the "
            "description in plain text (without quotes) for the following web page "
            "content:\n\n"
            "{content_html}"
        )

    @classmethod
    def image_title_prompt(cls):
        return (
            "Generate a title (in plain text, no longer than "
            "{max_length} characters, without quotes) for the following image: {image}"
        )

    @classmethod
    def image_description_prompt(cls):
        return (
            "Generate a description (in plain text, no longer than "
            "{max_length} characters) for the following image: {image}"
        )

    @classmethod
    def contextual_alt_text_prompt(cls):
        return """Generate an alt text (and only the text) for the following image:
{image}

Make the alt text relevant to the following content shown before the image:

---
{form_context_before}
---

and also relevant to the following content shown after the image:

---
{form_context_after}
---"""


class AgentSettingsMixin(models.Model):
    class ContentFeedbackContentType(models.TextChoices):
        TEXT = "text", _("Plain text")
        HTML = "html", _("HTML")

    page_title_prompt = models.TextField(
        blank=True,
        default=AgentPromptDefaults.page_title_prompt,
    )
    page_description_prompt = models.TextField(
        blank=True,
        default=AgentPromptDefaults.page_description_prompt,
    )
    image_title_prompt = models.TextField(
        blank=True,
        default=AgentPromptDefaults.image_title_prompt,
    )
    image_description_prompt = models.TextField(
        blank=True,
        default=AgentPromptDefaults.image_description_prompt,
    )
    contextual_alt_text_prompt = models.TextField(
        blank=True,
        default=AgentPromptDefaults.contextual_alt_text_prompt,
    )
    content_feedback_prompt = models.TextField(blank=True)
    content_feedback_content_type = models.CharField(
        max_length=64,
        choices=ContentFeedbackContentType.choices,
        default=ContentFeedbackContentType.HTML,
    )

    panels = [
        MultiFieldPanel(
            [
                FieldPanel(
                    "page_title_prompt",
                    heading=_("Page title prompt"),
                    help_text=_("Prompt to use for generating page titles."),
                ),
                FieldPanel(
                    "page_description_prompt",
                    heading=_("Page description prompt"),
                    help_text=_("Prompt to use for generating page descriptions."),
                ),
            ],
            heading=_("Page metadata"),
        ),
        MultiFieldPanel(
            [
                FieldPanel(
                    "image_title_prompt",
                    heading=_("Image title prompt"),
                    help_text=_(
                        "Prompt to use for generating image titles when uploading or "
                        "editing an image."
                    ),
                ),
                FieldPanel(
                    "image_description_prompt",
                    heading=_("Image description prompt"),
                    help_text=_(
                        "Prompt to use for generating image descriptions when "
                        "uploading or editing an image."
                    ),
                ),
                FieldPanel(
                    "contextual_alt_text_prompt",
                    heading=_("Contextual alt text prompt"),
                    help_text=_(
                        "Prompt to use for generating contextual image alt text "
                        "in the page editor."
                    ),
                ),
            ],
            heading=_("Images"),
        ),
        MultiFieldPanel(
            [
                FieldPanel(
                    "content_feedback_content_type",
                    heading=_("Content type"),
                    help_text=_("The format of the content to review."),
                ),
                FieldPanel(
                    "content_feedback_prompt",
                    heading=_("Prompt"),
                    help_text=_(
                        "Additional instructions to adjust the feedback given "
                        "by the agent."
                    ),
                ),
            ],
            heading=_("Content feedback"),
        ),
    ]

    class Meta:  # type: ignore
        abstract = True
        verbose_name = _("Agents")

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude)
        # Ensure these prompts have a value, falling back to defaults if needed.
        for field_name in [
            "page_title_prompt",
            "page_description_prompt",
            "image_title_prompt",
            "image_description_prompt",
            "contextual_alt_text_prompt",
        ]:
            field_value = getattr(self, field_name)
            if not field_value:
                setattr(self, field_name, getattr(AgentPromptDefaults, field_name)())


class AgentSettings(AgentSettingsMixin, BaseGenericSetting):
    class Meta(AgentSettingsMixin.Meta):  # type: ignore
        abstract = False

    def __str__(self):
        return str(self._meta.verbose_name)



from .base import get_llm_service
from .content_feedback import ContentFeedbackAgent
from .suggested_content import SuggestedContentAgent

__all__ = ["ContentFeedbackAgent", "SuggestedContentAgent", "get_llm_service"]



# Generated by Django 4.2.25 on 2025-10-09 08:05

from django.db import migrations, models

import wagtail_ai.models


class Migration(migrations.Migration):
    dependencies = [
        ("wagtail_ai", "0002_populate_default_prompts"),
    ]

    operations = [
        migrations.CreateModel(
            name="AgentSettings",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "page_title_prompt",
                    models.TextField(
                        blank=True,
                        default=wagtail_ai.models.AgentPromptDefaults.page_title_prompt,
                    ),
                ),
                (
                    "page_description_prompt",
                    models.TextField(
                        blank=True,
                        default=wagtail_ai.models.AgentPromptDefaults.page_description_prompt,
                    ),
                ),
                (
                    "image_title_prompt",
                    models.TextField(
                        blank=True,
                        default=wagtail_ai.models.AgentPromptDefaults.image_title_prompt,
                    ),
                ),
                (
                    "image_description_prompt",
                    models.TextField(
                        blank=True,
                        default=wagtail_ai.models.AgentPromptDefaults.image_description_prompt,
                    ),
                ),
                (
                    "contextual_alt_text_prompt",
                    models.TextField(
                        blank=True,
                        default=wagtail_ai.models.AgentPromptDefaults.contextual_alt_text_prompt,
                    ),
                ),
                ("content_feedback_prompt", models.TextField(blank=True)),
                (
                    "content_feedback_content_type",
                    models.CharField(
                        choices=[("text", "Plain text"), ("html", "HTML")],
                        default="html",
                        max_length=64,
                    ),
                ),
            ],
            options={
                "verbose_name": "Agents",
                "abstract": False,
            },
        ),
    ]



from string import Formatter
from typing import Any, Type, cast
from urllib.parse import SplitResult, urlsplit

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.files import File
from django.utils.translation import gettext as _
from wagtail.images.models import AbstractImage


class PromptContext(dict[str, Any]):
    """
    A dict-like object to interpolate a prompt string's placeholders with values.
    Provides a ``clean()`` method to validate that all placeholders can be handled.

    Validators can be registered with ``register_validator()`` to validate and
    transform a key's value or provide a fallback.

    A missing key without a registered validator is set as ``{key}`` so
    ``str.format_map()`` leaves it as-is.
    """

    validators: dict[str, Any] = {}

    @classmethod
    def register_validator(cls, name: str, validator):
        """
        Register a validator function for a given placeholder key.

        Validator functions should accept a single argument, the context
        instance, and return the value to use for the key. They can raise
        ``ValidationError`` if the key is required but invalid or missing from
        the context.
        """
        cls.validators[name] = validator

    def clean(self, prompt: str):
        """
        Validate that all placeholder keys in the prompt can be handled.
        """
        formatter = Formatter()
        placeholders = [name for _, name, *_ in formatter.parse(prompt) if name]
        for key in placeholders:
            if key in self.validators:
                # Provide the whole context to the validator to allow it to
                # access other keys if needed
                self[key] = self.validators[key](self)
            elif key not in self:
                self[key] = "{%s}" % key


def image_validator(context) -> File | SplitResult | None:
    """Fetch a Wagtail image by ID and return its rendition file."""
    if "image" not in context:
        raise ValidationError(_("The prompt requires an image, but none was provided."))
    image = context["image"]

    # Check if it's a data URL
    try:
        url: SplitResult = urlsplit(image)
    except (AttributeError, ValueError):
        pass
    else:
        if url.scheme == "data":
            if url.path and url.path.startswith("image/"):
                # Return the SplitResult instead of opening the URL,
                # as some backends (e.g. OpenAI) can handle data URLs directly.
                return url
            raise ValidationError(
                _("The provided data URL is not an image."),
                code="invalid",
            )

    from wagtail.images import get_image_model

    Image = cast(Type[AbstractImage], get_image_model())
    try:
        image = Image._default_manager.get(id=image)
    except Image.DoesNotExist as error:
        raise ValidationError(
            _("Could not find image with id %(image)s.") % {"image": image}
        ) from error
    wagtail_ai_settings = getattr(settings, "WAGTAIL_AI", {})
    rendition_filter = wagtail_ai_settings.get(
        "IMAGE_DESCRIPTION_RENDITION_FILTER", "max-800x600"
    )
    rendition = image.get_rendition(rendition_filter)
    image_file = rendition.file
    return image_file


PromptContext.register_validator("image", image_validator)
